import { Type } from "./types/generic_type";

/**
 * A Component can be thought of as a collection of data
 * which is attached to an entity.
 */
export abstract class Component {
    public readonly type: Type<Component>;
    public active: boolean;
    public entity_id: number;

    constructor(type: Type<Component>, entity_id: number) {
        this.type = type;
        this.entity_id = entity_id;
        this.active = true;
        if (this.on_initialize !== undefined) { this.on_initialize(); }
    }

    /**
     * @internal
     *
     * Activates the component, making it visible to engines..
     * Components are active by default when added to an entity.
     */
    public activate() {
        if (!this.active) {
            if (this.on_activate !== undefined) { this.on_activate(); }
            this.active = true;
        }
    }

    /**
     * @internal
     *
     * Deactivates the component, making it invisible to engines.
     */
    public deactivate() {
        if (this.active) {
            if (this.on_deactivate !== undefined) { this.on_deactivate(); }
            this.active = false;
        }
    }

    /**
     * Destroys the component and triggers the on_destroy callback.
     * Calling this function from outside the framework is a mistake.
     */
    public destroy() {
        if (this.on_destroy !== undefined) { this.on_destroy(); }
    }

    protected on_initialize?(): void;
    protected on_activate?(): void;
    protected on_deactivate?(): void;
    protected on_destroy?(): void;
}
