import { GCOptimizedList, GCOptimizedMap } from "tstl-gc-optimized-collections";
import { Component } from "./component";
import { DrawLayerManager } from "./draw_layer_manager";
import { ComponentUserDataOnly } from "./interfaces/component_user_data_only";
import { IDrawable } from "./interfaces/drawable";
import { Type } from "./types/generic_type";

export class ComponentManager {
    private component_type_to_active_components_map = new GCOptimizedMap<Type<Component>, GCOptimizedList<Component>>();
    private component_type_to_inactive_components_map = new GCOptimizedMap<Type<Component>, GCOptimizedList<Component>>();

    private components_to_activate = new GCOptimizedList<Component>();
    private components_to_deactivate = new GCOptimizedList<Component>();
    private components_to_remove = new GCOptimizedList<Component>();

    public create_component<TComponent extends Component>(
        ComponentType: Type<TComponent>,
        entity_id: number,
        draw_layer_manager: DrawLayerManager<IDrawable>,
        properties?: ComponentUserDataOnly<TComponent>,
    ): TComponent {
        const component = new ComponentType(ComponentType, entity_id, draw_layer_manager);
        if (!this.component_type_to_active_components_map.has(ComponentType)) {
            this.component_type_to_active_components_map.set(ComponentType, new GCOptimizedList<TComponent>());
            this.component_type_to_inactive_components_map.set(ComponentType, new GCOptimizedList<TComponent>());
        }

        this.mark_for_activation(component);

        // initializing the properties
        if (properties !== undefined) {
            for (const [key, value] of Object.entries(properties)) {
                (component as any)[key] = value;
            }
        }

        return component;
    }

    public get_components_by_component_type<TComponent extends Component>(
        ComponentType: Type<TComponent>,
    ): GCOptimizedList<Readonly<TComponent>> {
        if (this.component_type_to_active_components_map.has(ComponentType)) {
            return this.component_type_to_active_components_map.get(
                ComponentType,
            )! as GCOptimizedList<TComponent>;
        } else {
            return GCOptimizedList.Empty;
        }
    }

    public get_inactive_components_by_component_type<TComponent extends Component>(
        ComponentType: Type<TComponent>,
    ): GCOptimizedList<Readonly<TComponent>> {
        if (this.component_type_to_inactive_components_map.has(ComponentType)) {
            return this.component_type_to_inactive_components_map.get(ComponentType)! as GCOptimizedList<TComponent>;
        } else {
            return GCOptimizedList.Empty;
        }
    }

    public activate_components() {
        for (const component of this.components_to_activate.values()) {
            this.perform_activate_component(component);
        }
        this.components_to_activate.clear();
    }

    public deactivate_components() {
        for (const component of this.components_to_deactivate.values()) {
            this.perform_deactivate_component(component);
        }
        this.components_to_deactivate.clear();
    }

    public remove_components() {
        for (const component of this.components_to_remove.values()) {
            this.perform_remove_component(component);
        }
        this.components_to_remove.clear();
    }

    public mark_for_activation(component: Component) {
        this.components_to_activate.add(component);
    }

    public mark_for_deactivation(component: Component) {
        this.components_to_deactivate.add(component);
    }

    public mark_for_removal(component: Component) {
        this.components_to_remove.add(component);
    }

    protected perform_remove_component(component: Component) {
        this.component_type_to_active_components_map.get(component.type)!.deleteValue(component);
        this.component_type_to_inactive_components_map.get(component.type)!.deleteValue(component);
    }

    protected perform_deactivate_component(component: Component) {
        this.component_type_to_active_components_map.get(component.type)!.deleteValue(component);
        this.component_type_to_inactive_components_map.get(component.type)!.add(component);
    }

    protected perform_activate_component(component: Component) {
        this.component_type_to_inactive_components_map.get(component.type)!.deleteValue(component);
        this.component_type_to_active_components_map.get(component.type)!.add(component);
    }
}
