import { GCOptimizedList } from "tstl-gc-optimized-collections";
import { Component } from "./component";
import { ComponentManager } from "./component_manager";
import { Entity } from "./entity";
import { EntityCreator } from "./entity_creator";
import { EntityManager } from "./entity_manager";
import { EmitUndeclaredMessageError } from "./errors/emit_undeclared_message";
import { IllegalComponentMutationError } from "./errors/illegal_component_mutation";
import { ReadUndeclaredMessage } from "./errors/read_undeclared_message";
import { SingletonReadError } from "./errors/singleton_read";
import { Message } from "./message";
import { MessageManager } from "./message_manager";
import { ComponentMessage } from "./types/component_message";
import { EntityMessage } from "./types/entity_message";
import { Type } from "./types/generic_type";

/**
 * An Engine is the Encompass notion of an ECS System. Engines are responsible
 * for reading the game state and emitting messages, as well as mutating
 * Entities and Components.
 *
 * A particular kind of Component may be modified by only one Engine. If
 * you have two Engines which mutate the same Component type, an error
 * will be thrown.
 *
 * An Engine which Reads a particular Message type is guaranteed to run
 * after all Engines which Emit that particular Message type.
 */
export abstract class Engine {
    public readonly emit_message_types = new GCOptimizedList<Type<Message>>();
    public readonly read_message_types = new GCOptimizedList<Type<Message>>();
    public readonly mutate_component_types = new GCOptimizedList<Type<Component>>();

    private entity_manager: EntityManager;
    private entity_creator: EntityCreator;
    private component_manager: ComponentManager;
    private message_manager: MessageManager;

    constructor(
        entity_manager: EntityManager,
        entity_creator: EntityCreator,
        component_manager: ComponentManager,
        message_manager: MessageManager,
    ) {
        this.entity_manager = entity_manager;
        this.entity_creator = entity_creator;
        this.component_manager = component_manager;
        this.message_manager = message_manager;
    }

    public abstract update(dt: number): void;

    /**
     * Returns a newly instantiated Entity.
     */
    protected create_entity(): Entity {
        return this.entity_creator.create_entity();
    }

    /**
     * Emits a newly instantiated Message of MessageType.
     * The message can and should be modified after this method is called.
     *
     * @param MessageType
     * @throws When the MessageType is undeclared.
     */
    protected emit_message<TMessage extends Message>(
        MessageType: Type<TMessage>,
    ): TMessage {
        if (this.emit_message_types.hasValue(MessageType)) {
            return this.message_manager.create_message(MessageType);
        } else {
            const message_error = new EmitUndeclaredMessageError(this.constructor.name, MessageType);
            throw message_error.toString();
        }
    }

    /**
     * A convenience method for emitting messages that have a component property.
     * The message can and should be modified after this method is called.
     *
     * @param MessageType
     * @param component
     */
    protected emit_component_message<TMessage extends ComponentMessage>(
        MessageType: Type<TMessage>,
        component: Component,
    ): TMessage {
        const message = this.emit_message(MessageType);
        message.component = component;
        return message;
    }

    /**
     * A convenience method for emitting messages that have an entity property.
     * The message can and should be modified after this method is called.
     *
     * @param MessageType
     * @param entity
     */
    protected emit_entity_message<TMessage extends EntityMessage>(
        MessageType: Type<TMessage>,
        entity: Entity,
    ): TMessage {
        const message = this.emit_message(MessageType);
        message.entity = entity;
        return message;
    }

    /**
     * Returns the Entity instance with the given ID.
     *
     * @param entity_id
     */
    protected get_entity(entity_id: number): Entity | undefined {
        return this.entity_manager.get_entity(entity_id);
    }

    /**
     * Gets all active components of ComponentType.
     * The components are marked as readonly. For mutable components,
     * use get_mutable_components_by_type.
     *
     * @param ComponentType
     */
    protected read_components<TComponent extends Component>(
        ComponentType: Type<TComponent>,
    ): GCOptimizedList<Readonly<TComponent>> {
        return this.component_manager.get_components_by_component_type(ComponentType);
    }

    /**
     * A convenience method for the case of singleton Components.
     *
     * @param ComponentType
     * @throws If more than one component of ComponentType exists
     * @returns A component of type ComponentType
     */
    protected read_component<TComponent extends Component>(
        ComponentType: Type<TComponent>,
    ): Readonly<TComponent> | null {
        const components = this.component_manager.get_components_by_component_type(ComponentType);
        if (components.size === 0) {
            return null;
        } else if (components.size === 1) {
            return components.get(0)!;
        } else {
            const multiple_component_error = new SingletonReadError(this.constructor.name, ComponentType);
            throw multiple_component_error.toString();
        }
    }

    /**
     * Gets all active components of ComponentType.
     * The components are mutable, but the ComponentType must
     * be declared in mutates_component_types on the Engine
     * or an error will be thrown.
     *
     * @param ComponentType
     * @throws When the ComponentType is not declared.
     */
    protected read_components_mutable<TComponent extends Component>(
        ComponentType: Type<TComponent>,
    ): GCOptimizedList<TComponent> {
        if (this.mutate_component_types.hasValue(ComponentType)) {
            return this.component_manager.get_components_by_component_type(ComponentType);
        } else {
            const mutate_error = new IllegalComponentMutationError(this.constructor.name, ComponentType);
            throw mutate_error.toString();
        }
    }

    /**
     * A convenience method for the case of singleton Components.
     *
     * @param ComponentType
     * @throws When the ComponentType is not declared.
     * @throws When there is more than one Component.
     * @returns A component of type ComponentType
     *
     */
    protected read_component_mutable<TComponent extends Component>(
        ComponentType: Type<TComponent>,
    ): TComponent | null {
        if (this.mutate_component_types.hasValue(ComponentType)) {
            const components = this.component_manager.get_components_by_component_type(ComponentType);
            if (components.size === 0) {
                return null;
            } else if (components.size === 1) {
                return components.get(0)!;
            } else {
                const multiple_component_error = new SingletonReadError(this.constructor.name, ComponentType);
                throw multiple_component_error.toString();
            }
        } else {
            const mutate_error = new IllegalComponentMutationError(this.constructor.name, ComponentType);
            throw mutate_error.toString();
        }
    }

    /**
     * Makes a given component instance mutable.
     *
     * @param component
     * @throws When the component is not declared in mutate_component_types.
     */
    protected make_mutable<TComponent extends Component>(component: Readonly<TComponent>): TComponent {
        if (this.mutate_component_types.hasValue(component.type)) {
            return component;
        } else {
            const mutate_error = new IllegalComponentMutationError(this.constructor.name, component.type);
            throw mutate_error.toString();
        }
    }

    /**
     * Gets all inactive components of ComponentType.
     * The components are marked as readonly.
     *
     * @param ComponentType
     */
    protected read_inactive_components<TComponent extends Component>(
        ComponentType: Type<TComponent>,
    ): GCOptimizedList<Readonly<TComponent>> {
        return this.component_manager.get_inactive_components_by_component_type(ComponentType);
    }

    /**
     * Gets all messages of MessageType.
     *
     * @param MessageType
     * @throws When the message type is undeclared.
     */
    protected read_messages<TMessage extends Message>(
        MessageType: Type<TMessage>,
    ): GCOptimizedList<Readonly<TMessage>> {
        if (this.read_message_types.hasValue(MessageType)) {
            return this.message_manager.get_messages_by_message_type(MessageType);
        } else {
            const message_error = new ReadUndeclaredMessage(this.constructor.name, MessageType);
            throw message_error.toString();
        }
    }

    /**
     * Returns true if any messages of given type were emitted this frame.
     *
     * @param MessageType
     * @throws When the message type is undeclared.
     */
    protected some<TMessage extends Message>(
        MessageType: Type<TMessage>,
    ): boolean {
        if (this.read_message_types.hasValue(MessageType)) {
            return this.message_manager.get_messages_by_message_type(MessageType).size > 0;
        } else {
            const message_error = new ReadUndeclaredMessage(this.constructor.name, MessageType);
            throw message_error.toString();
        }
    }
}
