import { Component } from "../component";
import { Engine } from "../engine";
import { Entity } from "../entity";
import { IEntityTracker } from "../interfaces/entity_tracker";
import { Type } from "../types/generic_type";
import { EntityChecker } from "./utility/entity_checker";
import { EntityTracker } from "./utility/entity_tracker";

/**
 * A Detector is an Engine which is responsible
 * for reading the game world and producing Messages
 * in response.
 *
 * Detector are defined by the Component types they
 * track and the detect function they implement.
 *
 * It is an anti-pattern to modify Entities or
 * Components from inside a Detector.
 */
export abstract class Detector extends Engine implements IEntityTracker {
    public readonly component_types: Type<Component>[];

    private entity_tracker = new EntityTracker();

    public update(dt: number) {
        for (const id of this.entity_tracker.tracked_ids.entries()) {
            const entity = this.get_entity(id)!;
            this.detect(entity, dt);
        }
    }

    public check_and_track_entity(entity_id: number) {
        const entity = this.get_entity(entity_id)!;
        const should_track = EntityChecker.check_entity(entity, this.component_types);
        if (should_track) { this.entity_tracker.track_entity(entity_id); }
        return should_track;
    }

    public check_and_untrack_entity(entity_id: number) {
        const entity = this.get_entity(entity_id);
        const should_untrack = entity === undefined || !EntityChecker.check_entity(entity, this.component_types);
        if (should_untrack) {
            this.entity_tracker.untrack_entity(entity_id);
        }
        return should_untrack;
    }

    public is_tracking(id: number) {
        return this.entity_tracker.tracked_ids.has(id);
    }

    /**
     * Each World update, this callback runs for each
     * entity tracked by the Detector.
     *
     * @param entity An entity that is being tracked
     * by the Detector.
     */
    protected abstract detect(entity: Entity, dt: number): void;
}
