import { GCOptimizedList, GCOptimizedMap } from "tstl-gc-optimized-collections";
import { Component } from "../../component";
import { Engine } from "../../engine";
import { IllegalComponentMutationError } from "../../errors/illegal_component_mutation";
import { ComponentMessage } from "../../types/component_message";
import { Type } from "../../types/generic_type";
import { Mutable } from "../../types/mutable";

/**
 * A ComponentModifier is an Engine which is responsible
 * for consuming ComponentMessages and modifying the referenced
 * Component in response.
 *
 * ComponentModifiers are defined by their message_type
 * and the modify function they implement.
 *
 * It is an anti-pattern to read Entities or Components
 * within a ComponentModifier.
 */
export abstract class ComponentModifier extends Engine {
    private component_to_messages = new GCOptimizedMap<Readonly<Component>, GCOptimizedList<ComponentMessage>>();

    public update(dt: number) {
        this.retrieve_and_map_messages_by_component_instance();
        for (const [component, messages] of this.component_to_messages.entries()) {
            const entity = this.get_entity(component.entity_id);
            if (entity !== undefined && messages.size > 0) {
                if (this.mutate_component_types.hasValue(component.type)) {
                    this.modify(component, messages, dt);
                } else {
                    const mutation_error = new IllegalComponentMutationError(this.constructor.name, component.type);
                    throw mutation_error.toString();
                }
            }
        }
        this.clean();
    }

    protected abstract modify(
        component: Mutable<Component>,
        messages: GCOptimizedList<ComponentMessage>,
        dt: number,
    ): void;

    protected clean() {
        for (const [component, messages] of this.component_to_messages.entries()) {
            if (!this.get_entity(component.entity_id)) {
                this.component_to_messages.delete(component);
            }
            messages.clear();
        }
    }

    private retrieve_and_map_messages_by_component_instance() {
        const messages = this.read_messages(this.read_message_types.get(0)! as Type<ComponentMessage>);
        for (const message of messages.values()) {
            if (!this.component_to_messages.has(message.component)) {
                this.component_to_messages.set(message.component, new GCOptimizedList<ComponentMessage>());
            }
            this.component_to_messages.get(message.component)!.add(message);
        }
    }
}
