import { GCOptimizedList, GCOptimizedMap } from "tstl-gc-optimized-collections";
import { Engine } from "../../engine";
import { Entity } from "../../entity";
import { EntityMessage } from "../../types/entity_message";
import { Type } from "../../types/generic_type";

/**
 * An EntityModifier is an Engine which is responsible
 * for consuming EntityMessages and modifying Entities
 * in response.
 *
 * EntityModifiers are defined by their message_type
 * and the modify function they implement.
 *
 * It is an anti-pattern to read data from Entities
 * or Components inside an EntityModifier.
 */
export abstract class EntityModifier extends Engine {
    private entity_id_to_messages = new GCOptimizedMap<number, GCOptimizedList<EntityMessage>>();

    public update(dt: number) {
        this.retrieve_and_map_messages_by_entity_id();
        for (const [entity_id, entity_messages] of this.entity_id_to_messages.entries()) {
            const entity = this.get_entity(entity_id);
            if (entity !== undefined && entity_messages.size > 0) {
                this.modify(entity, entity_messages, dt);
            }
        }
        this.clean();
    }

    protected clean() {
        for (const [entity_id, messages] of this.entity_id_to_messages.entries()) {
            if (!this.get_entity(entity_id)) {
                this.entity_id_to_messages.delete(entity_id);
            }
            messages.clear();
        }
    }

    protected abstract modify(entity: Entity, messages: GCOptimizedList<EntityMessage>, dt: number): void;

    private retrieve_and_map_messages_by_entity_id() {
        const messages = this.read_messages(this.read_message_types.get(0)! as Type<EntityMessage>);
        for (const [_, message] of messages.entries()) {
            if (!this.entity_id_to_messages.has(message.entity.id)) {
                this.entity_id_to_messages.set(message.entity.id, new GCOptimizedList<EntityMessage>());
            }
            this.entity_id_to_messages.get(message.entity.id)!.add(message);
        }
    }
}
