import { GCOptimizedList, GCOptimizedMap, GCOptimizedSet } from "tstl-gc-optimized-collections";
import { Component } from "./component";
import { ComponentManager } from "./component_manager";
import { DrawLayerManager } from "./draw_layer_manager";
import { ComponentUserDataOnly } from "./interfaces/component_user_data_only";
import { IDrawable } from "./interfaces/drawable";
import { IEntityTracker } from "./interfaces/entity_tracker";
import { Type } from "./types/generic_type";

// TODO: refactor this stuff onto component and entity manager

/**
 * An Entity is composed of a unique internal ID
 * and a collection of Components.
 *
 * Entities do not have any implicit properties or
 * behavior. They are granted these by their collection
 * of Components.
 *
 * There is no limit to the amount of Components an
 * Entity can have, and Entities can have any number
 * of Components of a particular type.
 *
 * Entities are active by default and can be deactivated.
 * When deactivated, Entities are still tracked by
 * Engines, but the Engines temporarily ignore them.
 *
 * Note that the activation status of an Entity
 * does not affect the activation status of its Components.
 *
 * Defining logic on an Entity is an anti-pattern.
 */
export class Entity {
    public readonly id: number;

    private _component_types = new GCOptimizedSet<Type<Component>>();
    private component_manager: ComponentManager;

    private component_type_to_component_bag: GCOptimizedMap<Type<Component>, GCOptimizedList<Component>>;
    private component_type_to_active_components: GCOptimizedMap<Type<Component>, GCOptimizedList<Component>>;

    private dirty_entities_added: GCOptimizedList<Entity>;
    private dirty_entities_removed: GCOptimizedList<Entity>;
    private entities_marked_for_destroy: GCOptimizedList<Entity>;

    private draw_layer_manager: DrawLayerManager<IDrawable>;

    get component_types(): GCOptimizedSet<Type<Component>> { return this._component_types; }

    constructor(
        id: number,
        dirty_entities_added: GCOptimizedList<Entity>,
        dirty_entities_removed: GCOptimizedList<Entity>,
        entities_marked_for_destroy: GCOptimizedList<Entity>,
        draw_layer_manager: DrawLayerManager<IDrawable>,
        component_manager: ComponentManager,
    ) {
        this.id = id;
        this.dirty_entities_added = dirty_entities_added;
        this.dirty_entities_removed = dirty_entities_removed;
        this.entities_marked_for_destroy = entities_marked_for_destroy;
        this.draw_layer_manager = draw_layer_manager;
        this.component_manager = component_manager;

        this.component_type_to_component_bag = new GCOptimizedMap<Type<Component>, GCOptimizedList<Component>>();
        this.component_type_to_active_components = new GCOptimizedMap<Type<Component>, GCOptimizedList<Component>>();
    }

    /**
     * Instantiates a Component of the given subtype
     * and adds it to the Entity.
     *
     * @param ComponentType {Type<TComponent>}
     * @param properties {ComponentUserDataOnly<TComponent>}
     */
    public add_component<TComponent extends Component>(
        ComponentType: Type<TComponent>,
        properties?: ComponentUserDataOnly<TComponent>,
    ): TComponent {
        const component = this.component_manager.create_component(
            ComponentType,
            this.id,
            this.draw_layer_manager,
            properties,
        );

        if (!this.component_type_to_component_bag.has(ComponentType)) {
            const component_bag = new GCOptimizedList<TComponent>();
            const active_component_bag = new GCOptimizedList<TComponent>();
            this.component_type_to_component_bag.set(ComponentType, component_bag);
            this.component_type_to_active_components.set(ComponentType, active_component_bag);
            component_bag.add(component);
            active_component_bag.add(component);
        } else {
            this.component_type_to_component_bag.get(ComponentType)!.add(component);
            this.component_type_to_active_components.get(ComponentType)!.add(component);
        }

        this._component_types.add(ComponentType);
        this.dirty_entities_added.add(this);

        return component;
    }

    /**
     * This is a convenience method for the common case
     * of having a single component of a certain type on
     * an Entity.
     *
     * NOTE: Only returns an active Component.
     *
     * NOTE: Can return undefined if no Components of the given type exist.
     *
     * @param constructor
     * @returns An arbitrary readonly component of given type, or null if none exist.
     */
    public get_component<TComponent extends Component>(constructor: Type<TComponent>): Readonly<TComponent> {
        return this.get_components(constructor).get(0)!;
    }

    /**
     * Gets all the active Components of a given type that exist
     * on the Entity.
     *
     * @param ComponentType
     * @returns A GCOptimizedList containing readonly references to each
     * component of the given Component type, or an empty GCOptimizedList if none exist.
     */
    public get_components<TComponent extends Component>(ComponentType: Type<TComponent>): GCOptimizedList<Readonly<TComponent>> {
        if (this.component_type_to_active_components.has(ComponentType)) {
            return this.component_type_to_active_components.get(ComponentType)! as GCOptimizedList<TComponent>;
        } else {
            return GCOptimizedList.Empty;
        }
    }

    /**
     * Gets all the Components of a given type that exist on the Entity,
     * including inactive Components.
     *
     * @param ComponentType
     * @returns A GCOptimizedList containing references to each
     * component of the given Component type, or an empty GCOptimizedList if none exist.
     */
    public get_components_including_inactive<TComponent extends Component>(ComponentType: Type<TComponent>): GCOptimizedList<Readonly<TComponent>> {
        if (this.component_type_to_component_bag.has(ComponentType)) {
            return this.component_type_to_component_bag.get(ComponentType)! as GCOptimizedList<TComponent>;
        } else {
            return GCOptimizedList.Empty;
        }
    }

    /**
     * Checks if the Entity has a particular active Component of ComponentType.
     *
     * @param ComponentType
     * @returns True if the given Component type is on the Entity, otherwise false.
     */
    public has_component<TComponent extends Component>(ComponentType: Type<TComponent>): boolean {
        return this.component_type_to_active_components.has(ComponentType) &&
              !this.component_type_to_active_components.get(ComponentType)!.empty();
    }

    /**
     * Checks if the Entity has a particular Component of ComponentType, including inactive components.
     *
     * @param ComponentType
     * @returns True if the given Component type is on the Entity, otherwise false.
     */
    public has_component_including_inactive<TComponent extends Component>(ComponentType: Type<TComponent>): boolean {
        return this.component_type_to_component_bag.has(ComponentType) &&
              !this.component_type_to_component_bag.get(ComponentType)!.empty();
    }

    /**
     * Removes a specific Component instance from
     * the Entity.
     *
     * @param component
     */
    public remove_component(component: Component) {
        const component_bag = this.component_type_to_component_bag.get(component.type)!;
        const active_component_bag = this.component_type_to_active_components.get(component.type)!;
        component_bag.deleteValue(component);
        active_component_bag.deleteValue(component);
        this.dirty_entities_removed.add(this);
        this.component_manager.mark_for_removal(component);
        component.destroy();
        if (component_bag.empty()) { this._component_types.delete(component.type); }
    }

    public remove_all_components() {
        for (const [type, component_list] of this.component_type_to_component_bag.entries()) {
            for (const component of component_list.values()) {
                this.component_manager.mark_for_removal(component);
                component.destroy();
            }

            component_list.clear();
        }

        for (const [type, component_list] of this.component_type_to_active_components.entries()) {
            component_list.clear();
        }

        this._component_types.clear();
        this.dirty_entities_removed.add(this);
    }

    /**
     * Activates the component, making it tracked by engines.
     * Does nothing if the component is already active.
     *
     * NOTE: Components are active by default when added to an entity.
     *
     * @param component The component to activate.
     */
    public activate_component(component: Component) {
        if (!component.active) {
            component.activate();
            this.component_type_to_active_components.get(component.type)!.add(component);
            this.dirty_entities_added.add(this);
            this.component_manager.mark_for_activation(component);
        }
    }

    /**
     * Deactivates the component, making it untracked by engines.
     * Does nothing if the component is already inactive.
     *
     * @param component The component to deactivate.
     */
    public deactivate_component(component: Component) {
        if (component.active) {
            component.deactivate();
            const active_component_bag = this.component_type_to_active_components.get(component.type)!;
            const index = active_component_bag.indexOf(component);
            if (index !== null) {
                active_component_bag.delete(index);
            }
            this.dirty_entities_removed.add(this);
            this.component_manager.mark_for_deactivation(component);
        }
    }

    /**
     * Marks the Entity to be destroyed. Note
     * that the Entity will exist until the
     * end of the World update.
     */
    public destroy() {
        this.entities_marked_for_destroy.add(this);
    }
}
