import { GCOptimizedList, GCOptimizedMap, GCOptimizedSet } from "tstl-gc-optimized-collections";
import { Component } from "./component";
import { Entity } from "./entity";
import { IDrawableEntityTracker } from "./interfaces/drawable_entity_tracker";
import { IEntityTracker } from "./interfaces/entity_tracker";
import { Type } from "./types/generic_type";

export class EntityManager {
    private entities_with_added_components: GCOptimizedList<Entity>;
    private entities_with_removed_components: GCOptimizedList<Entity>;
    private entities_marked_for_destroy: GCOptimizedList<Entity>;

    private entities: GCOptimizedList<Entity>;
    private id_to_entity: GCOptimizedMap<number, Entity>;

    private entity_to_entity_trackers_map = new GCOptimizedMap<Entity, GCOptimizedSet<IEntityTracker>>();
    private component_type_to_entity_trackers_map = new GCOptimizedMap<Type<Component>, GCOptimizedList<IEntityTracker>>();

    constructor(
        entities: GCOptimizedList<Entity>,
        id_to_entity: GCOptimizedMap<number, Entity>,
        entities_with_added_components: GCOptimizedList<Entity>,
        entities_with_removed_components: GCOptimizedList<Entity>,
        entities_marked_for_destroy: GCOptimizedList<Entity>,
    ) {
        this.entities = entities;
        this.id_to_entity = id_to_entity;
        this.entities_with_added_components = entities_with_added_components;
        this.entities_with_removed_components = entities_with_removed_components;
        this.entities_marked_for_destroy = entities_marked_for_destroy;
    }

    public check_entities_with_added_components() {
        for (const entity of this.entities_with_added_components.values()) {
            this.check_and_register_entity(entity);
        }
        this.entities_with_added_components.clear();
    }

    public check_entities_with_removed_components() {
        for (const entity of this.entities_with_removed_components.values()) {
            if (this.entity_to_entity_trackers_map.has(entity)) {
                for (const engine of this.entity_to_entity_trackers_map.get(entity)!.entries()) {
                    if (this.is_entity_tracker(engine)) {
                        engine.check_and_untrack_entity(entity.id);
                    }
                }
            }
        }
        this.entities_with_removed_components.clear();
    }

    public register_entity_tracker(entity_tracker: IEntityTracker) {
        for (const component_type of entity_tracker.component_types) {
            const entity_tracker_set = this.get_or_create_entity_tracker_set(component_type);
            entity_tracker_set.add(entity_tracker);
        }
    }

    public register_drawable_entity_tracker(drawable_entity_tracker: IDrawableEntityTracker) {
        this.register_entity_tracker(drawable_entity_tracker);
        const entity_tracker_set = this.get_or_create_entity_tracker_set(drawable_entity_tracker.draw_component_type);
        entity_tracker_set.add(drawable_entity_tracker);
    }

    public destroy_marked_entities() {
        for (const entity of this.entities_marked_for_destroy.values()) {
            entity.remove_all_components();

            if (this.entity_to_entity_trackers_map.has(entity)) {
                for (const tracker of this.entity_to_entity_trackers_map.get(entity)!.entries()) {
                    tracker.check_and_untrack_entity(entity.id);
                }
            }

            this.entity_to_entity_trackers_map.delete(entity);

            this.entities.deleteValue(entity);
            this.id_to_entity.delete(entity.id);
        }

        this.entities_marked_for_destroy.clear();
    }

    public get_entity(id: number): Entity | undefined {
        return this.id_to_entity.get(id);
    }

    private is_entity_tracker(engine: any): engine is IEntityTracker {
        return (engine as IEntityTracker).
        check_and_untrack_entity !== undefined;
    }

    private get_or_create_entity_tracker_set(component_type: Type<Component>) {
        if (!this.component_type_to_entity_trackers_map.has(component_type)) {
            this.component_type_to_entity_trackers_map.set(component_type, new GCOptimizedList<IEntityTracker>());
        }
        return this.component_type_to_entity_trackers_map.get(component_type)!;
    }

    private check_and_register_entity(entity: Entity) {
        for (const component_type of entity.component_types.entries()) {
            if (this.component_type_to_entity_trackers_map.has(component_type)) {
                for (const entity_tracker of this.component_type_to_entity_trackers_map.get(component_type)!.values()) {
                    if (entity_tracker.check_and_track_entity(entity.id)) {
                        if (!this.entity_to_entity_trackers_map.has(entity)) {
                            this.entity_to_entity_trackers_map.set(entity, new GCOptimizedSet<IEntityTracker>());
                        }
                        this.entity_to_entity_trackers_map.get(entity)!.add(entity_tracker);
                    }
                }
            }
        }
    }
}
