import { Graph } from "./graph";
import { PriorityQueue } from "./priority_queue";

enum SearchSymbol {
    start,
    finish,
}

export class DirectedGraph<T> extends Graph<T> {
    public add_edge(u: T, v: T) {
        if (this.vertex_exists(u) && this.vertex_exists(v)) {
            this._edges.get(u)!.push(v);
        }
    }

    public add_edges(...edges: [T, T][]) {
        for (const edge of edges) {
            this.add_edge(edge[0], edge[1]);
        }
    }

    public node_dfs(): Map<T, Map<SearchSymbol, number>> {
        const discovered = new Set<T>();
        let time = 0;
        const output = new Map<T, Map<SearchSymbol, number>>();

        for (const vertex of this.vertices) {
            output.set(vertex, new Map<SearchSymbol, number>());
        }

        const dfs_helper = (u: T) => {
            discovered.add(u);
            time += 1;
            output.get(u)!.set(SearchSymbol.start, time);

            for (const neighbor of this.neighbors(u)) {
                if (!discovered.has(neighbor)) {
                    dfs_helper(neighbor);
                }
            }

            time += 1;
            output.get(u)!.set(SearchSymbol.finish, time);
        };

        for (const vertex of this.vertices) {
            if (!discovered.has(vertex)) {
                dfs_helper(vertex);
            }
        }

        return output;
    }

    public topological_sort(): T[] {
        const dfs = this.node_dfs();
        const priority_queue = new PriorityQueue<T>();
        for (const [key, value] of dfs) {
            priority_queue.add(key, value.get(SearchSymbol.finish)!);
        }
        return priority_queue.to_array().reverse();
    }

    public clone(): DirectedGraph<T> {
        const clone = new DirectedGraph<T>();
        clone.add_vertices(...this.vertices);

        for (const v of this.vertices) {
            for (const n of this.neighbors(v)) {
                clone.add_edge(v, n);
            }
        }

        return clone;
    }

    public cyclic(): boolean {
        const sccs = this.strongly_connected_components();
        for (const scc of sccs) {
            if (scc.length > 1) {
                return true;
            }
        }
        return false;
    }

    public acyclic(): boolean {
        return !this.cyclic();
    }

    /**
     * Uses Tarjan's algorithm with Nuutila's modifications. Nonrecursive.
     */
    public strongly_connected_components(): T[][] {
        const preorder = new Map<T, number>();
        const lowlink = new Map<T, number>();
        const scc_found = new Map<T, boolean>();
        const scc_queue = [];

        const result: T[][] = [];

        let i = 0; // preorder counter

        for (const source of this.vertices) {
            if (!scc_found.has(source)) {
                const queue = [source];
                while (queue.length > 0) {
                    const v = queue[queue.length - 1];
                    if (!preorder.has(v)) {
                        i += 1;
                        preorder.set(v, i);
                    }
                    let done = true;
                    const v_neighbors = this.neighbors(v);
                    for (const w of v_neighbors) {
                        if (!preorder.has(w)) {
                            queue.push(w);
                            done = false;
                            break;
                        }
                    }
                    if (done) {
                        lowlink.set(v, preorder.get(v)!);
                        for (const w of v_neighbors) {
                            if (!scc_found.has(w)) {
                                if (preorder.get(w)! > preorder.get(v)!) {
                                    lowlink.set(v, Math.min(lowlink.get(v)!, lowlink.get(w)!));
                                } else {
                                    lowlink.set(v, Math.min(lowlink.get(v)!, preorder.get(w)!));
                                }
                            }
                        }
                        queue.pop();
                        if (lowlink.get(v) === preorder.get(v)) {
                            scc_found.set(v, true);
                            const scc: T[] = [];
                            scc.push(v);
                            while (scc_queue.length > 0 && preorder.get(scc_queue[scc_queue.length - 1])! > preorder.get(v)!) {
                                const k = scc_queue.pop()!;
                                scc_found.set(k, true);
                                scc.push(k);
                            }
                            result.push(scc);
                        } else {
                            scc_queue.push(v);
                        }
                    }
                }
            }
        }

        return result;
    }

    /**
     * Uses Johnson's cycle finding algorithm.
     */
    public simple_cycles() {
        const unblock = (thisnode: T, blocked: Set<T>, B: Map<T, Set<T>>) => {
            const stack = [thisnode];
            while (stack.length > 0) {
                const node = stack.pop()!;
                if (blocked.has(node)) {
                    blocked.delete(node);
                    if (B.has(node)) {
                        for (const n of B.get(node)!) {
                            if (stack.indexOf(n) === -1) {
                                stack.push(n);
                            }
                        }
                        B.get(node)!.clear();
                    }
                }
            }
        };

        const result: T[][] = [];
        const subG = this.clone();
        const sccs = [];
        for (const scc of this.strongly_connected_components()) {
            sccs.push(scc);
        }

        while (sccs.length > 0) {
            const scc = sccs.pop()!;
            const start_node = scc.pop()!;
            const path = [start_node];
            const blocked = new Set<T>();
            const closed = new Set<T>();
            blocked.add(start_node);
            const B = new Map<T, Set<T>>();
            const stack: [T, T[]][] = [[start_node, subG.neighbors(start_node)]];

            while (stack.length > 0) {
                const [thisnode, neighbors] = stack[stack.length - 1];

                if (neighbors.length > 0) {
                    const next_node = neighbors.pop()!;

                    if (next_node === start_node) {
                        const result_path: T[] = [];
                        for (const v of path) {
                            result_path.push(v);
                        }
                        result.push(result_path);
                        for (const v of path) {
                            closed.add(v);
                        }
                    } else if (!blocked.has(next_node)) {
                        path.push(next_node);
                        stack.push([next_node, subG.neighbors(next_node)]);
                        closed.delete(next_node);
                        blocked.add(next_node);
                        continue;
                    }
                }

                if (neighbors.length === 0) {
                    if (closed.has(thisnode)) {
                        unblock(thisnode, blocked, B);
                    } else {
                        for (const neighbor of subG.neighbors(thisnode)) {
                            if (!B.has(neighbor)) {
                                B.set(neighbor, new Set<T>());
                            }
                            B.get(neighbor)!.add(thisnode);
                        }
                    }
                    stack.pop();
                    path.pop();
                }
            }
            subG.remove_vertex(start_node);
            const H = subG.subgraph(scc);
            const H_sccs = H.strongly_connected_components();
            for (const H_scc of H_sccs) {
                sccs.push(H_scc);
            }
        }

        // remove duplicates
        const eqSet = (as: Set<T>, bs: Set<T>) => {
            if (as.size !== bs.size) { return false; }
            for (const a of as) {
                if (!bs.has(a)) { return false; }
            }
            return true;
        };

        const deduplicate = (cycles: T[][]) => {
            const uniques: Set<T>[] = [];
            for (const array of cycles) {
                const set = new Set(array);
                let add = true;
                for (const unique of uniques) {
                    if (eqSet(set, unique)) {
                        add = false;
                    }
                }
                if (add) {
                    uniques.push(set);
                }
            }
            return uniques.map((set) => {
                const array = [];
                for (const value of set) {
                    array.push(value);
                }
                return array;
            });
        };

        return deduplicate(result);
    }

    public subgraph(vertices: T[]) {
        const subgraph = new DirectedGraph<T>();
        subgraph.add_vertices(...vertices);
        for (const v of this.vertices) {
            if (vertices.indexOf(v) !== -1) {
                const neighbors = this.neighbors(v);
                for (const u of neighbors) {
                    if (vertices.indexOf(u) !== -1) {
                        subgraph.add_edge(v, u);
                    }
                }
            }
        }
        return subgraph;
    }

    public remove_edge(edge: [T, T]): void {
        const neighbors = this._edges.get(edge[0])!;
        neighbors.splice(neighbors.indexOf(edge[1])!, 1);
    }
}
