import { GCOptimizedList, GCOptimizedMap, GCOptimizedSet } from "tstl-gc-optimized-collections";
import { Component } from "./component";
import { Entity } from "./entity";
import { Message } from "./message";
import { MessagePool } from "./message_pool";
import { MessageQuery } from "./message_query";
import { ComponentMessage } from "./types/component_message";
import { EntityMessage } from "./types/entity_message";
import { Type } from "./types/generic_type";

export class MessageManager {
    public message_query = new MessageQuery(); // delegate through?
    private messages = new GCOptimizedSet<Message>();

    private message_pools = new GCOptimizedMap<Type<Message>, MessagePool<Message>>();

    private message_type_to_messages_map = new GCOptimizedMap<Type<Message>, GCOptimizedList<Message>>();

    public create_component_message<TComponentMessage extends ComponentMessage>(
            ComponentMessageType: Type<TComponentMessage>,
            component: Component,
        ): TComponentMessage {
        const component_message = this.create_or_get_message_pool(ComponentMessageType).obtain()! as TComponentMessage;
        component_message.component = component;
        this.messages.add(component_message);

        this.cache_message_by_message_type(component_message);
        return component_message;
    }

    public create_entity_message<TEntityMessage extends EntityMessage>(
        EntityMessageType: Type<TEntityMessage>,
        entity: Entity,
    ): TEntityMessage {
        const entity_message = this.create_or_get_message_pool(EntityMessageType).obtain()! as TEntityMessage;
        entity_message.entity = entity;
        this.messages.add(entity_message);
        this.cache_message_by_message_type(entity_message);
        return entity_message;
    }

    public create_message<TMessage extends Message>(
        MessageType: Type<TMessage>,
    ): TMessage {
        const message = this.create_or_get_message_pool(MessageType).obtain()! as TMessage;
        this.messages.add(message);
        this.cache_message_by_message_type(message);
        return message;
    }

    public clear_messages() {
        for (const [_, message_set] of this.message_type_to_messages_map.entries()) {
            message_set.clear();
        }

        for (const message of this.messages.entries()) {
            this.message_pools.get(message.type)!.free(message);
        }

        this.messages.clear();
    }

    public get_messages_by_message_type<TMessage extends Message>(MessageType: Type<TMessage>): GCOptimizedList<TMessage> {
        if (this.message_type_to_messages_map.has(MessageType)) {
            return this.message_type_to_messages_map.get(MessageType)! as GCOptimizedList<TMessage>;
        } else {
            return GCOptimizedList.Empty;
        }
    }

    private cache_message_by_message_type(message: Message) {
        if (!this.message_type_to_messages_map.get(message.type)) {
            this.message_type_to_messages_map.set(message.type, new GCOptimizedList<Message>());
        }
        this.message_type_to_messages_map.get(message.type)!.add(message);
    }

    private create_or_get_message_pool<TMessage extends Message>(MessageType: Type<TMessage>): MessagePool<Message> {
        if (!this.message_pools.has(MessageType)) {
            this.message_pools.set(MessageType, new MessagePool<Message>(MessageType));
        }
        return this.message_pools.get(MessageType)!;
    }
}
