import { GCOptimizedList } from "tstl-gc-optimized-collections";
import { Type } from "./types/generic_type";

export class ObjectPool<T> {
    get size(): number { return this._size; }
    get inactive_objects(): GCOptimizedList<T> { return this._inactive_objects; }
    get active_objects(): GCOptimizedList<T> { return this._active_objects; }

    public static OverflowBehaviorFallible = () => null;

    public static OverflowBehaviorCycle = <T>(pool: ObjectPool<T>) => {
        return pool.active_objects.shift()!;
    }

    public static OverflowBehaviorExpand = <T>(pool: ObjectPool<T>) => {
        pool.generate(pool.size);
        pool._size *= 2;
        return pool._inactive_objects.pop()!;
    }

    public static DefaultDestructor = <T>(object: T) => {
        // nullify the properties
        for (const key in object) {
            if (key !== "_type" && key !== "type") {
                (object as any)[key] = undefined;
            }
        }
    }

    public overflow_behavior: (pool: ObjectPool<T>) => T | null;
    private _size: number;
    private _inactive_objects: GCOptimizedList<T>;
    private _active_objects: GCOptimizedList<T>;
    private constructor_function: new(constructor: Type<T>) => T;
    private destructor_function: (object: T) => void;
    private reference_object: T;

    constructor(
        object_constructor: new(constructor: Type<T>) => T,
        object_destructor: (destructor: T) => void = ObjectPool.DefaultDestructor,
        size = 16,
        overflow_behavior: (pool: ObjectPool<T>) => T | null = ObjectPool.OverflowBehaviorExpand,
    ) {
        this._size = size;
        this._inactive_objects = new GCOptimizedList<T>();
        this._active_objects = new GCOptimizedList<T>();

        this.reference_object = new object_constructor(object_constructor);

        this.constructor_function = object_constructor;
        this.destructor_function = object_destructor;
        this.overflow_behavior = overflow_behavior;
        this.generate(size);
    }

    public obtain(): T | null {
        const object = this.get_inactive_object();
        if (object !== null) {
            for (const key in object) {
                if (key !== "_type" && key !== "type") {
                    (object as any)[key] = (this.reference_object as any)[key];
                }
            }
            this._active_objects.add(object);
        }
        return object;
    }

    public free(instance: T) {
        if (this.active_objects.deleteValue(instance)) {
            this.inactive_objects.add(instance);
            this.destructor_function(instance);
        } else {
            throw("cannot free object which is not from pool");
        }
    }

    public free_all() {
        for (const [_, object] of this.active_objects.entries()) {
            this.free(object);
        }
    }

    private generate(amount: number) {
        let i = 0;
        while (i < amount) {
            const object = new this.constructor_function(this.constructor_function);
            this._inactive_objects.add(object);
            i += 1;
        }
    }

    private get_inactive_object(): T | null {
        if (this._inactive_objects.size > 0) {
            return this._inactive_objects.pop()!;
        } else {
            return this.overflow_behavior(this);
        }
    }
}
