import { GCOptimizedList } from "tstl-gc-optimized-collections";
import { Component } from "./component";
import { ComponentManager } from "./component_manager";
import { EntityManager } from "./entity_manager";
import { SingletonReadError } from "./errors/singleton_read";
import { Type } from "./types/generic_type";

export abstract class Renderer {
    private entity_manager: EntityManager;
    private component_manager: ComponentManager;

    constructor(
        entity_manager: EntityManager,
        component_manager: ComponentManager,
    ) {
        this.entity_manager = entity_manager;
        this.component_manager = component_manager;
    }

    protected get_entity(entity_id: number) {
        return this.entity_manager.get_entity(entity_id);
    }

    /**
     * Gets all active components of ComponentType.
     * The components are marked as readonly. For mutable components,
     * use get_mutable_components_by_type.
     *
     * @param ComponentType
     */
    protected read_components<TComponent extends Component>(
        ComponentType: Type<TComponent>,
    ): GCOptimizedList<Readonly<TComponent>> {
        return this.component_manager.get_components_by_component_type(ComponentType);
    }

    /**
     * A convenience method for the case of singleton Components.
     *
     * @param ComponentType
     * @throws If more than one component of ComponentType exists
     * @returns A component of type ComponentType
     */
    protected read_component<TComponent extends Component>(
        ComponentType: Type<TComponent>,
    ): Readonly<TComponent> | null {
        const components = this.component_manager.get_components_by_component_type(ComponentType);
        if (components.size === 0) {
            return null;
        } else if (components.size === 1) {
            return components.get(0)!;
        } else {
            const multiple_component_error = new SingletonReadError(this.constructor.name, ComponentType);
            throw multiple_component_error.toString();
        }
    }
}
