import * as jprof from "encompass-jprof";
import { GCOptimizedList } from "tstl-gc-optimized-collections";
import { ComponentManager } from "./component_manager";
import { Engine } from "./engine";
import { EntityCreator } from "./entity_creator";
import { EntityManager } from "./entity_manager";
import { MessageManager } from "./message_manager";
import { RenderManager } from "./render_manager";

/**
 * A World is the glue that holds all the framework
 * elements together.
 *
 * It is a mistake to instantiate elements of encompass
 * separately from a World. These elements should be
 * created either through appropriate World functions,
 * or by Engines.
 *
 * It is perfectly valid to use multiple Worlds in
 * the same game. For example, one World could
 * drive the main menu, and another could drive
 * the game logic.
 */
export class World {
    public engine_order: GCOptimizedList<Engine>;

    private component_manager: ComponentManager;
    private entity_creator: EntityCreator;
    private entity_manager: EntityManager;
    private message_manager: MessageManager;
    private render_manager: RenderManager;

    constructor(
        component_manager: ComponentManager,
        entity_creator: EntityCreator,
        entity_manager: EntityManager,
        message_manager: MessageManager,
        render_manager: RenderManager,
    ) {
        this.component_manager = component_manager;
        this.entity_creator = entity_creator;
        this.entity_manager = entity_manager;
        this.message_manager = message_manager;
        this.render_manager = render_manager;
    }

    /**
     * Updates the simulation based on given delta time,
     * advancing the simulation by one frame.
     *
     * @param dt Delta time
     */
    public update(dt: number) {
        jprof.push("frame");
        jprof.push("World.update");

        jprof.push("process");
        this.process(dt);
        jprof.pop("process");
        jprof.push("World.message_manager.clear_messages");
        this.message_manager.clear_messages();
        jprof.pop("World.message_manager.clear_messages");

        jprof.push("World.destroy_marked_entities");
        this.destroy_marked_entities();
        jprof.pop("World.destroy_marked_entities");

        jprof.push("World.entity_manager.check_entities_with_added_components");
        this.entity_manager.check_entities_with_added_components();
        jprof.pop("World.entity_manager.check_entities_with_added_components");
        jprof.push("World.entity_manager.check_entities_with_removed_components");
        this.entity_manager.check_entities_with_removed_components();
        jprof.pop("World.entity_manager.check_entities_with_removed_components");

        jprof.push("World.component_manager.activate_components");
        this.component_manager.activate_components();
        jprof.pop("World.component_manager.activate_components");
        jprof.push("World.component_manager.deactivate_components");
        this.component_manager.deactivate_components();
        jprof.pop("World.component_manager.deactivate_components");
        jprof.push("World.component_manager.remove_components");
        this.component_manager.remove_components();
        jprof.pop("World.component_manager.remove_components");

        jprof.pop("World.update");
    }

    /**
     * Calls the render methods of all Renderers in
     * the World ordered by the layers on DrawComponents.
     */
    public draw() {
        jprof.push("World.draw");
        this.render_manager.draw();
        jprof.pop("World.draw");
        jprof.pop("frame");
    }

    private destroy_marked_entities() {
        this.entity_manager.destroy_marked_entities();
    }

    private process(dt: number) {
        for (const engine of this.engine_order.values()) {
            jprof.push(engine.constructor.name + ".update");
            engine.update(dt);
            jprof.pop(engine.constructor.name + ".update");
        }
    }
}
