import { GCOptimizedList, GCOptimizedMap, GCOptimizedSet } from "tstl-gc-optimized-collections";
import { Component } from "./component";
import { ComponentManager } from "./component_manager";
import { Engine } from "./engine";
import { Detector } from "./engines/detector";
import { Entity } from "./entity";
import { EntityCreator } from "./entity_creator";
import { EntityManager } from "./entity_manager";
import { DetectorWithoutComponentTypesError } from "./errors/detector_without_component_types";
import { EngineCycleError } from "./errors/engine_cycle";
import { EngineMutationConflictError } from "./errors/engine_mutation_conflict";
import { DirectedGraph } from "./graphs/directed_graph";
import { Message } from "./message";
import { MessageManager } from "./message_manager";
import { RenderManager } from "./render_manager";
import { Renderer } from "./renderer";
import { TypeGuard } from "./type_guards";
import { ComponentMessage } from "./types/component_message";
import { EntityMessage } from "./types/entity_message";
import { Type } from "./types/generic_type";
import { World } from "./world";

/**
 * A WorldBuilder is used to build a World out of Engines and Renderers.
 *
 * The WorldBuilder enforces certain rules about Engine structure.
 * It is forbidden to have messages create cycles between Engines,
 * and no Component may be mutated by more than one Engine.
 *
 * The WorldBuilder uses Engines and their Message read/emit information
 * to determine a valid ordering of the Engines,
 * which is given to the World.
 */
export class WorldBuilder {
    private engines = new GCOptimizedList<Engine>();
    private renderers = new GCOptimizedList<Renderer>();

    private engine_graph = new DirectedGraph<Engine>();

    private message_type_to_emitters = new GCOptimizedMap<Type<Message>, GCOptimizedSet<Engine>>();
    private message_type_to_readers = new GCOptimizedMap<Type<Message>, GCOptimizedSet<Engine>>();

    private component_manager: ComponentManager;
    private entity_creator: EntityCreator;
    private entity_manager: EntityManager;
    private message_manager: MessageManager;
    private render_manager: RenderManager;

    private world: World;

    constructor() {
        const entities = new GCOptimizedList<Entity>();
        const id_to_entity = new GCOptimizedMap<number, Entity>();

        const entities_with_added_components = new GCOptimizedList<Entity>();
        const entities_with_removed_components = new GCOptimizedList<Entity>();
        const entities_marked_for_destroy = new GCOptimizedList<Entity>();

        this.render_manager = new RenderManager(id_to_entity);

        this.component_manager = new ComponentManager();

        this.entity_creator = new EntityCreator(
            entities,
            id_to_entity,
            entities_with_added_components,
            entities_with_removed_components,
            entities_marked_for_destroy,
            this.component_manager,
            this.render_manager.draw_layer_manager,
        );

        this.entity_manager = new EntityManager(
            entities,
            id_to_entity,
            entities_with_added_components,
            entities_with_removed_components,
            entities_marked_for_destroy,
        );

        this.message_manager = new MessageManager();

        this.world = new World(
            this.component_manager,
            this.entity_creator,
            this.entity_manager,
            this.message_manager,
            this.render_manager,
        );
    }

    /**
     * Instantiates a new Entity.
     */
    public create_entity(): Entity {
        return this.entity_creator.create_entity();
    }

    /**
     * Creates a new message.
     * @param MessageType
     */
    public emit_message<TMessage extends Message>(
        MessageType: Type<TMessage>,
    ): TMessage {
        return this.message_manager.create_message(MessageType);
    }

    /**
     * Creates a new component message to be consumed
     * immediately upon the next World update.
     *
     * @param ComponentMessageType
     * @param component
     */
    public emit_component_message<TMessage extends ComponentMessage>(
        ComponentMessageType: Type<TMessage>,
        component: Component,
    ): TMessage {
        return this.message_manager.create_component_message(ComponentMessageType, component);
    }

    /**
     * Creates a new EntityMessage to be consumed
     * immediately upon the next World update.
     *
     * @param EntityMessageType
     * @param entity
     */
    public emit_entity_message<TMessage extends EntityMessage>(
        EntityMessageType: Type<TMessage>,
        entity: Entity,
    ): TMessage {
        return this.message_manager.create_entity_message(EntityMessageType, entity);
    }

    /**
     * Adds an engine to the WorldBuilder.
     *
     * @param EngineType
     */
    public add_engine<TEngine extends Engine>(
        EngineType: new(
            entity_manager: EntityManager,
            entity_creator: EntityCreator,
            component_manager: ComponentManager,
            message_manager: MessageManager,
        ) => TEngine,
    ) {
        const engine = new EngineType(
            this.entity_manager,
            this.entity_creator,
            this.component_manager,
            this.message_manager,
        );

        this.engines.add(engine);
        this.engine_graph.add_vertex(engine);

        if (engine instanceof Detector) {
            this.entity_manager.register_entity_tracker(engine);

            if (!engine.component_types) {
                const detect_error = new DetectorWithoutComponentTypesError(this.constructor.name);
                throw detect_error.toString();
            }
        }

        for (const emit_message_type of engine.emit_message_types.values()) {
            if (!this.message_type_to_emitters.has(emit_message_type)) {
                this.message_type_to_emitters.set(emit_message_type, new GCOptimizedSet<Engine>());
            }
            this.message_type_to_emitters.get(emit_message_type)!.add(engine);

            if (this.message_type_to_readers.has(emit_message_type)) {
                for (const reader of this.message_type_to_readers.get(emit_message_type)!.entries()) {
                    this.engine_graph.add_edge(engine, reader);
                }
            }
        }

        for (const read_message_type of engine.read_message_types.values()) {
            if (!this.message_type_to_readers.has(read_message_type)) {
                this.message_type_to_readers.set(read_message_type, new GCOptimizedSet<Engine>());
            }
            this.message_type_to_readers.get(read_message_type)!.add(engine);

            if (this.message_type_to_emitters.has(read_message_type)) {
                for (const emitter of this.message_type_to_emitters.get(read_message_type)!.entries()) {
                    this.engine_graph.add_edge(emitter, engine);
                }
            }
        }

        return engine;
    }

    /**
     * Adds a new Renderer engine to the WorldBuilder.
     *
     * @param RendererType
     */
    public add_renderer<TRenderer extends Renderer>(
        RendererType: new(
            entity_manager: EntityManager,
            component_manager: ComponentManager,
        ) => TRenderer,
    ) {
        const renderer = new RendererType(this.entity_manager, this.component_manager);
        this.renderers.add(renderer);
        if (TypeGuard.renderer_is_entity_renderer(renderer)) {
            this.render_manager.register_entity_renderer(renderer);
            this.entity_manager.register_drawable_entity_tracker(renderer);
        } else if (TypeGuard.renderer_is_general_renderer(renderer)) {
            this.render_manager.draw_layer_manager.register_object_with_layer(renderer, renderer.layer);
        }
        return renderer;
    }

    /**
     * Call this when you are done adding Engines and Renderers.
     *
     * @throws {EngineCycleError} When the Engine emit/reads produce a cycle between Engines.
     * @throws {EngineMutationConflictError} When multiple Engines mutate the same Component type.
     *
     * @returns {World} An instantiated World with a valid Engine ordering.
     */

    public build(): World {
        if (this.engine_graph.cyclic()) {
            const engine_cycle_error = new EngineCycleError(this.engine_graph.simple_cycles());
            throw engine_cycle_error.toString();
        }

        const mutated_component_types = new Set<Type<Component>>();
        const duplicate_mutations: Type<Component>[] = [];
        const component_to_engines_map = new Map<Type<Component>, Engine[]>();

        for (const engine of this.engines.values()) {
            for (const mutated_component_type of engine.mutate_component_types.values()) {
                if (mutated_component_types.has(mutated_component_type)) {
                    duplicate_mutations.push(mutated_component_type);
                } else {
                    mutated_component_types.add(mutated_component_type);
                }

                if (!component_to_engines_map.has(mutated_component_type)) {
                    component_to_engines_map.set(mutated_component_type, []);
                }
                component_to_engines_map.get(mutated_component_type)!.push(engine);
            }
        }

        if (duplicate_mutations.length > 0) {
            const engine_mutation_conflict_error = new EngineMutationConflictError(duplicate_mutations, component_to_engines_map);
            throw engine_mutation_conflict_error.toString();
        }

        const gc_optimized_engine_order = new GCOptimizedList<Engine>();
        for (const engine of this.engine_graph.topological_sort()) {
            gc_optimized_engine_order.add(engine);
        }

        this.world.engine_order = gc_optimized_engine_order;

        this.entity_manager.check_entities_with_added_components();
        this.entity_manager.check_entities_with_removed_components();
        this.component_manager.activate_components();
        this.component_manager.deactivate_components();
        this.component_manager.remove_components();
        return this.world;
    }
}
